package com.rathravane.drumlin.app.accounts;

import java.util.Collection;
import java.util.logging.Logger;

import com.rathravane.drumlin.app.accounts.DrumlinAccountPersistence.tagInfo;
import com.rathravane.till.data.oneWayHasher;
import com.rathravane.till.data.sha1HmacSigner;
import com.rathravane.till.data.uniqueStringGenerator;
import com.rathravane.till.logging.rrLogSetup;
import com.rathravane.till.time.clock;

/**
 * <p>A user and account manager. Because Drumlin does not include a model,
 * your application must provide an implementation of DrumlinAccountPersistence.</p>
 * 
 * <p>No requirements are made on the interface used for the U and A generic classes.
 * They represent Users and Accounts.</p>
 * 
 * <p>Accounts contain users, but it's okay for an account to contain just a single
 * user, and in that case, the account ID and user ID could be the same.</p>
 * 
 * <p>Note that user ids must be unique across ALL accounts -- the
 * account does not provide a scope over user IDs. So, accounts A and B may not both
 * contain a user called C.</p>
 * 
 * @author peter
 *
 * @param <U>
 * @param <A>
 */
public class DrumlinAccountMgr<U,A>
{
	/**
	 * Instantiate an account manager given the app's account factory interface.
	 * 
	 * @param acctDb
	 */
	public DrumlinAccountMgr ( DrumlinAccountPersistence<U,A> acctDb )
	{
		fAccountDb = acctDb;
	}

	/**
	 * Close the account manager.
	 */
	public void close ()
	{
		fAccountDb.close ();
	}

	/**
	 * return true if the account exists
	 * @param acctKey
	 * @return true if the account exists
	 * @throws msModelException
	 */
	public boolean accountExists ( String acctKey ) throws DrumlinAccountsException
	{
		return fAccountDb.accountExists ( acctKey );
	}

	/**
	 * load an account with the given key or throw if it doesn't exist
	 * @param acctKey
	 * @return the account object
	 * @throws DrumlinAccountsException 
	 */
	public A loadAccount ( String acctKey ) throws DrumlinAccountsException
	{
		return fAccountDb.loadAccount ( acctKey );
	}

	/**
	 * create an account with the given key
	 * @param acctKey
	 * @return the account object
	 * @throws DrumlinAccountsException 
	 * @throws msModelException
	 */
	public A createAccount ( String acctKey ) throws DrumlinAccountsException
	{
		return fAccountDb.createAccount ( acctKey );
	}

	/**
	 * Return true if a user by the given id exists. 
	 * @param userId
	 * @return true if the user exists
	 * @throws DrumlinAccountsException
	 */
	public boolean userExists ( String userId ) throws DrumlinAccountsException
	{
		return fAccountDb.userExists ( userId );
	}

	/**
	 * Load a user record from the account db. Note this is not an authentication operation.
	 * 
	 * @param userId
	 * @return a user record, or null
	 * @throws DrumlinAccountsException
	 */
	public U loadUser ( String userId ) throws DrumlinAccountsException
	{
		return fAccountDb.loadUser ( userId );
	}

	/**
	 * authenticate a user with user id and password
	 * @param userId
	 * @param password
	 * @return the user record, or null if the authentication failed.
	 * @throws DrumlinAccountsException when a load/lookup operation fails
	 */
	public U authenticateUserWithPassword ( String userId, String password ) throws DrumlinAccountsException
	{
		final U u = fAccountDb.loadUser ( userId );
		if ( u == null )
		{
			log.info ( "AUTH_FAIL: no such user " + userId + "." );
			return null;
		}

		final String acctId = fAccountDb.getAccountIdForUser ( userId );
		if ( acctId == null )
		{
			log.warning ( "AUTH_FAIL: no account ID for user " + userId + "." );
			return null;
		}

		final A a = fAccountDb.loadAccount ( acctId );
		if ( a == null )
		{
			log.warning ( "AUTH_FAIL: no such account " + acctId + ", referenced by " + userId + "." );
			return null;
		}
	
		U result = null;
		if ( !fAccountDb.isUserEnabled ( userId ) )
		{
			log.info ( "AUTH_FAIL: " + userId + "; user is disabled." );
		}
		else if ( !isPasswordValid ( userId, password ) )
		{
			log.info ( "AUTH_FAIL: " + userId + "; incorrect password." );
		}
		else if ( !fAccountDb.isAccountEnabled ( acctId ) )
		{
			log.info ( "AUTH_FAIL: " + userId + " ; account is disabled." );
		}
		else if ( isConfirmationOverdue ( userId ) )
		{
			log.info ( "AUTH_FAIL: " + userId + " ; user confirmation grace period expired." );
		}
		else
		{
			result = u;
		}
		return result;
	}

	/**
	 * Create a user given an account ID, user ID, and email address. Note that no password is
	 * associated with the account, so login is impossible until the password is set.
	 * 
	 * @param acctId
	 * @param userId
	 * @param userEmail
	 * @return the user object created in the account database, or null if the creation failed
	 * @throws DrumlinAccountsException
	 */
	public U createUser ( String acctId, String userId, String userEmail ) throws DrumlinAccountsException
	{
		if ( !fAccountDb.accountExists ( acctId ) ) throw new DrumlinAccountsException ( "Account does not exist." );
		if ( fAccountDb.userExists ( userId ) ) throw new DrumlinAccountsException ( "User exists." );

		return fAccountDb.createUser ( acctId, userId, userEmail );
	}

	/**
	 * Delete a user given an account ID and user ID. (Generally, it's probably better to just disable the user id...)
	 * 
	 * @param acctId
	 * @param userId
	 */
	public void deleteUser ( String acctId, String userId ) throws DrumlinAccountsException
	{
		if ( fAccountDb.accountExists ( acctId ) && fAccountDb.userExists ( userId ) )
		{
			fAccountDb.deleteUser ( userId );
		}
	}
	
	/**
	 * Require that the user is confirmed by a particular point in time. If confirmation does not
	 * occur, the user account is disabled.
	 * 
	 * A user is always in one of the following states:<br>
	 * 
	 * 		no confirmation required<br>
	 *		confirmation required, still in grace period<br>
	 *		confirmation required, outside of grace period<br>
	 *		confirmed<br>
	 * 
	 * @param u
	 * @param secondsUntilExpire
	 * @param nonsense
	 * @return a tag identifying the confirmation request (send this in an email, for example)
	 * @throws DrumlinAccountsException 
	 */
	public String requireUserConfirmation ( String userKey, long secondsUntilExpire, String nonsense ) throws DrumlinAccountsException
	{
		String tag = null;
		if ( fAccountDb.userExists ( userKey ) )
		{
			final DrumlinAccountPersistence.tagInfo ti = fAccountDb.getRequestTagByUser ( userKey, kTagType_ConfirmUser );
			if ( ti != null )
			{
				fAccountDb.deleteRequestByTag ( ti.fTag, kTagType_ConfirmUser );
			}

			tag = uniqueStringGenerator.createUrlKey ( nonsense );
			final long expiresAtSeconds = ( clock.now() / 1000 ) + secondsUntilExpire;

			fAccountDb.storeRequestTag ( userKey, kTagType_ConfirmUser, tag, expiresAtSeconds );
		}
		return tag;
	}

	/**
	 * Load and confirm a user record.
	 * @param userId
	 * @throws DrumlinAccountsException
	 */
	public void confirmUser ( String userKey ) throws DrumlinAccountsException
	{
		if ( fAccountDb.userExists ( userKey ) )
		{
			final DrumlinAccountPersistence.tagInfo ti = fAccountDb.getRequestTagByUser ( userKey, kTagType_ConfirmUser );
			if ( ti != null )
			{
				fAccountDb.deleteRequestByTag ( ti.fTag, kTagType_ConfirmUser );
				fAccountDb.deleteRequestTagOnUser ( userKey, ti.fTag, kTagType_ConfirmUser );
			}
		}
	}

	/**
	 * Confirm a user based on a confirmation tag.
	 * @param tag
	 * @throws DrumlinAccountsException 
	 */
	public boolean confirmUserWithTag ( String requestTag ) throws DrumlinAccountsException
	{
		final tagInfo ti = fAccountDb.getRequestTagByTag ( requestTag, kTagType_ConfirmUser );
		if ( ti == null ) return false; 	// ignore the request

		final long nowEpochTime = clock.now() / 1000;
		if ( ti.fExpiresAtEpochTime < nowEpochTime ) return false;	// out of date

		confirmUser ( ti.fUserId );

		fAccountDb.deleteRequestByTag ( requestTag, kTagType_PasswordReset );

		return true;
	}

	/**
	 * Enable/Disable a user.
	 * @param userId
	 * @param enabled
	 * @throws DrumlinAccountsException
	 */
	public void enableUser ( String userId, boolean enabled ) throws DrumlinAccountsException
	{
		fAccountDb.enableUser ( userId, enabled );
	}
	
	/**
	 * Enable/Disable an account. A disabled account prevents any of its users from logging in.
	 * @param acctId
	 * @param enabled
	 * @throws DrumlinAccountsException
	 */
	public void enableAccount ( String acctId, boolean enabled ) throws DrumlinAccountsException
	{
		fAccountDb.enableAccount ( acctId, enabled );
	}

	/**
	 * Set a user's password. Passwords are salted and hashed before being sent to the accounts
	 * database. Salts are created per user using the current time, a randomly generated UUID,
	 * and the application's signature (provided by the account db). The password provided is
	 * combined with some static data and the salt to generate a string that's hashed. The
	 * generated salt and the hashed password are then passed to the account db for storage.
	 * 
	 * @param u
	 * @param pwd
	 * @throws DrumlinAccountsException 
	 */
	public void setPassword ( String user, String pwd ) throws DrumlinAccountsException
	{
		final String nonsense = fAccountDb.getAppSignature();
		final String salt = generateSalt ( nonsense );
		final String hashedPassword = passwdHash ( pwd, salt );
		fAccountDb.storePasswordInfo ( user, salt, hashedPassword );
	}

	/**
	 * Request a password reset request that expires. The return value is a unique tag for the request
	 * that must be presented to complete the transaction. The password is NOT changed until the
	 * request is completed.
	 * 
	 * @param userKey
	 * @param secondsUntilExpire
	 * @param nonsense Arbitrary string to use in generating the returned reset tag.
	 * @return a unique reset request tag
	 * @throws DrumlinAccountsException 
	 * @throws msItemDoesNotExistException
	 * @throws unexpectedException
	 */
	public String requestPasswordReset ( String userKey, long secondsUntilExpire, String nonsense ) throws DrumlinAccountsException
	{
		String tag = null;
		if ( fAccountDb.userExists ( userKey ) )
		{
			final DrumlinAccountPersistence.tagInfo ti = fAccountDb.getRequestTagByUser ( userKey, kTagType_PasswordReset );
			if ( ti != null )
			{
				fAccountDb.deleteRequestByTag ( ti.fTag, kTagType_PasswordReset );
			}

			tag = uniqueStringGenerator.createUrlKey ( nonsense );
			final long expiresAtSeconds = ( clock.now() / 1000 ) + secondsUntilExpire;

			fAccountDb.storeRequestTag ( userKey, kTagType_PasswordReset, tag, expiresAtSeconds );
		}
		return tag;
	}

	private static final String kTagType_PasswordReset = "passwordReset";
	private static final String kTagType_ConfirmUser = "confirmUser";
	
	/**
	 * Complete the password reset using the new password. This also cancels any
	 * pending account confirmation, because it presumably went through the same
	 * email round-trip mechanism.
	 * 
	 * @param requestTag
	 * @param newPassword
	 * @throws DrumlinAccountsException 
	 * @throws unexpectedException
	 * @return the user id, or null on failure
	 */
	public String completePasswordReset ( String requestTag, String newPassword ) throws DrumlinAccountsException
	{
		final tagInfo ti = fAccountDb.getRequestTagByTag ( requestTag, kTagType_PasswordReset );
		if ( ti == null ) return null; 	// ignore the request

		String result = null;

		final long nowEpochTime = clock.now() / 1000;
		if ( ti.fExpiresAtEpochTime >= nowEpochTime )
		{
			setPassword ( ti.fUserId, newPassword );
			confirmUser ( ti.fUserId );
			result = ti.fUserId;
		}

		fAccountDb.deleteRequestByTag ( requestTag, kTagType_PasswordReset );
		fAccountDb.deleteRequestTagOnUser ( ti.fUserId, requestTag, kTagType_PasswordReset );

		return result;
	}

	/**
	 * Set a general purpose value on the user record.
	 * @param userId
	 * @param key
	 * @param value
	 * @throws DrumlinAccountsException 
	 */
	public void setUserValue ( String userId, String key, String value ) throws DrumlinAccountsException
	{
		fAccountDb.setUserValue ( userId, key, value );
	}

	/**
	 * Retrieve a general purpose value from the user record.
	 * @param userId
	 * @param key
	 * @return the stored value, or null
	 * @throws DrumlinAccountsException 
	 */
	public String getUserValue ( String userId, String key ) throws DrumlinAccountsException
	{
		return fAccountDb.getUserValue ( userId, key );
	}

	/**
	 * Remove a general purpose value from the user record.
	 * @param userId
	 * @param key
	 */
	public void removeUservalue ( String userId, String key ) throws DrumlinAccountsException
	{
		fAccountDb.removeUservalue ( userId, key );
	}

	/**
	 * Load an API key record.
	 * @param apiKey
	 * @return an API key or null
	 * @throws DrumlinAccountsException
	 */
	public DrumlinApiKey loadApiKey ( String apiKey ) throws DrumlinAccountsException
	{
		return fAccountDb.loadApiKey ( apiKey );
	}

	/**
	 * Load all API keys for a user.
	 * @param userId
	 * @return a collection of 0 or more API keys
	 * @throws DrumlinAccountsException
	 */
	public Collection<DrumlinApiKey> loadApiKeysForUser ( String userId ) throws DrumlinAccountsException
	{
		return fAccountDb.loadApiKeysForUser ( userId );
	}

	/**
	 * Authenticate a request using an api (public) key, a string of content that's signed, and the
	 * signature.
	 * 
	 * @param apiKey
	 * @param signedContent
	 * @param signature
	 * @return an API key record, which will indicate whether its an account or user access, or null
	 * @throws DrumlinAccountsException 
	 */
	public DrumlinApiKey authenticateViaApiKey ( String apiKey, String signedContent, String signature ) throws DrumlinAccountsException
	{
		DrumlinApiKey result = null;

		final DrumlinApiKey key = fAccountDb.loadApiKey ( apiKey );
		if ( key != null )
		{
			// use private key to sign content
			final String expectedSignature = sha1HmacSigner.sign ( signedContent, key.getSecret () );
			log.finer ( "expecting [" + expectedSignature + "]; received [" + signature + "]. signed content [" + signedContent + "]." );

			// compare
			if ( expectedSignature.equals ( signature ) )
			{
				result = key;
			}
			// else: not signed correctly
		}
		else
		{
			log.info ( "AUTH_FAIL: No API key [" + apiKey + "]" );
		}

		return result;
	}

	/**
	 * Create an API key for the given account.
	 * @param acct
	 * @return an API key
	 * @throws DrumlinAccountsException 
	 */
	public DrumlinApiKey createAcctApiKey ( String acctId ) throws DrumlinAccountsException
	{
		final String appSig = fAccountDb.getAppSignature();
		final String newApiKey = generateKey ( 16, appSig );
		final String newApiSecret = generateKey ( 24, appSig );
		return establishAcctApiKey ( acctId, newApiKey, newApiSecret );
	}

	/**
	 * Create an API key for the given user.
	 * @param user
	 * @return an API key
	 * @throws DrumlinAccountsException 
	 */
	public DrumlinApiKey createUserApiKey ( String user ) throws DrumlinAccountsException
	{
		final String appSig = fAccountDb.getAppSignature();
		final String newApiKey = generateKey ( 16, appSig );
		final String newApiSecret = generateKey ( 24, appSig );
		return establishUserApiKey ( user, newApiKey, newApiSecret );
	}

	/**
	 * Delete a user's api key.
	 * @param user
	 * @param apiKey
	 * @throws DrumlinAccountsException
	 */
	public void deleteUserApiKey ( String user, String apiKey ) throws DrumlinAccountsException
	{
		final DrumlinApiKey key = fAccountDb.loadApiKey ( apiKey );
		if ( key != null && key.getAccessType ().equals ( DrumlinApiKey.type.USER ) && key.getUser ().equals ( user ) )
		{
			fAccountDb.deleteApiKey ( key );
		}
	}

	/**
	 * This is provided for internal use, not use from user-facing code.
	 * @param acct
	 * @param apiKey
	 * @param newApiSecret
	 * @return
	 * @throws DrumlinAccountsException 
	 * @throws msModelException
	 */
	public DrumlinApiKey establishAcctApiKey ( final String acctId, final String apiKey, final String newApiSecret ) throws DrumlinAccountsException
	{
		DrumlinApiKey key = null;

		// if the API key exists (mainly used for test), load it and update it
		key = loadApiKey ( apiKey );
		if ( key != null && key.getAccount () != null && key.getAccount().equals ( acctId ) )
		{
			// update the existing key
			log.info ( "API key " + apiKey + " already exists for account [" + acctId + "]. Resetting API secret." );
			fAccountDb.setApiSecret ( apiKey, newApiSecret );
		}
		else
		{
			// create a new key
			key = new DrumlinApiKey ()
			{
				@Override
				public String getKey () { return apiKey; }

				@Override
				public String getSecret () { return newApiSecret; }

				@Override
				public DrumlinApiKey.type getAccessType () { return DrumlinApiKey.type.ACCOUNT; }

				@Override
				public String getAccount () { return acctId; }

				@Override
				public String getUser () { return null; }
			};
			fAccountDb.storeApiKey ( key );
		}

		return key;
	}

	/**
	 * This is provided for internal use, not use from user-facing code.
	 * @param acct
	 * @param apiKey
	 * @param newApiSecret
	 * @return
	 * @throws DrumlinAccountsException 
	 * @throws msModelException
	 */
	public DrumlinApiKey establishUserApiKey ( final String userId, final String apiKey, final String newApiSecret ) throws DrumlinAccountsException
	{
		DrumlinApiKey key = null;

		// if the API key exists (mainly used for test), load it and update it
		key = loadApiKey ( apiKey );
		if ( key != null && key.getUser () != null && key.getUser().equals ( userId ) )
		{
			// update the existing key
			log.info ( "API key " + apiKey + " already exists for user [" + userId + "]. Resetting API secret." );
			fAccountDb.setApiSecret ( apiKey, newApiSecret );
		}
		else
		{
			// create a new key
			key = new DrumlinApiKey ()
			{
				@Override
				public String getKey () { return apiKey; }

				@Override
				public String getSecret () { return newApiSecret; }

				@Override
				public DrumlinApiKey.type getAccessType () { return DrumlinApiKey.type.USER; }

				@Override
				public String getAccount () { return null; }

				@Override
				public String getUser () { return userId; }
			};
			fAccountDb.storeApiKey ( key );
		}

		return key;
	}

	private final DrumlinAccountPersistence<U,A> fAccountDb;

	private boolean isPasswordValid ( String userId, String pwd ) throws DrumlinAccountsException
	{
		boolean result = false;
		
		final String[] pwdData = fAccountDb.getPasswordInfo ( userId );
		
		final String hash = passwdHash ( pwd, pwdData[0] );
		final String stored = pwdData[1];
		if ( stored != null )
		{
			result = hash.equals ( stored );
		}
		return result;
	}

	private boolean isConfirmationOverdue ( String userKey ) throws DrumlinAccountsException
	{
		// it's overdue if there's an expired request out there.
		if ( fAccountDb.userExists ( userKey ) )
		{
			final DrumlinAccountPersistence.tagInfo ti = fAccountDb.getRequestTagByUser ( userKey, kTagType_ConfirmUser );
			if ( ti != null )
			{
				final long now = clock.now() / 1000;
				return ti.fExpiresAtEpochTime < now;
			}
		}
		return false;
	}

	private static String passwdHash ( String p, String salt )
	{
		return oneWayHasher.hash ( p, salt );
	}

	private static String generateSalt ( String nonsense )
	{
		return generateKey ( 8, nonsense );
	}

	private static String kKeyChars = "ABCDEFGHJIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";
	private static String generateKey ( int length, String nonsense  )
	{
		return uniqueStringGenerator.createKeyUsingAlphabet ( nonsense, kKeyChars, length );
	}

	private static final Logger log = rrLogSetup.getLog ( DrumlinAccountMgr.class );
}
