package com.rathravane.drumlin.app.accounts;

import java.util.Collection;

/**
 * An interface to the application's model/persistence system for user and account management.
 * While this interface could be de(re?)-composed into a few more interfaces (user, account),
 * it's left as a single set of methods to implement to keep it self-contained, and to prevent
 * the interface structure from creating trouble in terms of managing state across concurrent
 * instances.
 * 
 * @author peter
 *
 * @param <U>
 * @param <A>
 */
public interface DrumlinAccountPersistence<U,A>
{
	/**
	 * The application needs a constant signature string. This can be pretty much anything.
	 * It's used to construct various hashes.
	 * 
	 * @return an arbitrary string, but always the same string.
	 */
	String getAppSignature ();

	/**
	 * Called when the account manager is closed.
	 */
	void close ();
	
	/**
	 * Create a User and add it to the given account. The user should be created with
	 * unusable authentication information. The account manager will provide authentication
	 * information in a separate call.
	 * 
	 * @param acct
	 * @param uid
	 * @param userEmail
	 * @param addlData
	 * @return a user
	 * @throws DrumlinAccountsException
	 */
	U createUser ( String acctId, String uid, String userEmail ) throws DrumlinAccountsException;

	/**
	 * Delete a user, removing it from its account.
	 * @param uid
	 * @throws DrumlinAccountsException
	 */
	void deleteUser ( String uid ) throws DrumlinAccountsException;

	/**
	 * Load a user. (This loads the user record; it does not deal with authenticating
	 * a user.)
	 * 
	 * @param uid
	 * @return a user or null
	 * @throws DrumlinAccountsException
	 */
	U loadUser ( String uid ) throws DrumlinAccountsException;

	/**
	 * Does the user exist?
	 * @param uid
	 * @return
	 * @throws DrumlinAccountsException
	 */
	boolean userExists ( String uid ) throws DrumlinAccountsException;

	void enableUser ( String id, boolean enabled ) throws DrumlinAccountsException;

	boolean isUserEnabled ( String userId ) throws DrumlinAccountsException;

	String getAccountIdForUser ( String userId ) throws DrumlinAccountsException;

	void storePasswordInfo ( String userId, String salt, String hashedPassword ) throws DrumlinAccountsException;

	/**
	 * Return the salt and hashed password stored for the given user.
	 * @param userId
	 * @return
	 * @throws DrumlinAccountsException
	 */
	String[] getPasswordInfo ( String userId ) throws DrumlinAccountsException;

	/**
	 * Tags are chunks of data associated with a user ID, used for things like password reset via email confirmation. Tags can expire.  
	 */
	class tagInfo
	{
		public tagInfo ( String tag, String userId, long expiresAtEpochTime ) { fTag=tag; fUserId=userId; fExpiresAtEpochTime=expiresAtEpochTime; }
		public final String fTag;
		public final String fUserId;
		public final long fExpiresAtEpochTime;
	}
	
	/**
	 * Store a tag. Note that the tag may be requested with userId+tagType or tagId+tagType.
	 * @param userId
	 * @param tagType
	 * @param tagId
	 * @param expiresAtEpochTime
	 * @throws DrumlinAccountsException
	 */
	void storeRequestTag ( String userId, String tagType, String tagId, long expiresAtEpochTime ) throws DrumlinAccountsException;

	/**
	 * Get a tag given the tagId and tagType provided to storeRequestTag. Return null if
	 * the record is not found. Implementations should ignore the expiration time on a tag
	 * and return even expired tag records. 
	 *  
	 * @param tag
	 * @param tagType
	 * @return the tag or null
	 * @throws DrumlinAccountsException
	 */
	tagInfo getRequestTagByTag ( String tag, String tagType ) throws DrumlinAccountsException;

	/**
	 * Get a tag given the userId and tagType provided to storeRequestTag. Return null if
	 * the record is not found. Implementations should ignore the expiration time on a tag
	 * and return even expired tag records.
	 * 
	 * @param userId
	 * @param tagType
	 * @return the tag or null
	 * @throws DrumlinAccountsException
	 */
	tagInfo getRequestTagByUser ( String userId, String tagType ) throws DrumlinAccountsException;

	/**
	 * Delete a tag given the tagId and tagType. (The tagType is provided for consistency with
	 * storeRequestTag; your implementation may not require it.) Implementations should
	 * take care to ensure that a getRequestTagByUser call with the given tagType will return
	 * null after this call.
	 * 
	 * @param tag
	 * @param tagType
	 * @throws DrumlinAccountsException
	 */
	void deleteRequestByTag ( String tag, String tagType ) throws DrumlinAccountsException;

	/**
	 * Delete a tag given the userId, tagId, and tagType. 
	 * 
	 * @param tag
	 * @param tagType
	 * @throws DrumlinAccountsException
	 */
	void deleteRequestTagOnUser ( String user, String tag, String tagType ) throws DrumlinAccountsException;
	
	A createAccount ( String id ) throws DrumlinAccountsException;
	A loadAccount ( String id ) throws DrumlinAccountsException;
	boolean accountExists ( String id ) throws DrumlinAccountsException;

	void enableAccount ( String id, boolean enabled ) throws DrumlinAccountsException;
	boolean isAccountEnabled ( String id ) throws DrumlinAccountsException;

	/**
	 * Given an API key, load the associated API key and account information.
	 * @param apiKey
	 * @return an api key instance
	 * @throws DrumlinAccountsException
	 */
	DrumlinApiKey loadApiKey ( String apiKey ) throws DrumlinAccountsException;
	Collection<DrumlinApiKey> loadApiKeysForUser ( String userId ) throws DrumlinAccountsException;

	void storeApiKey ( DrumlinApiKey apiKey ) throws DrumlinAccountsException;
	void deleteApiKey ( DrumlinApiKey apiKey ) throws DrumlinAccountsException;

	void setApiSecret ( String apiKey, String apiSecret ) throws DrumlinAccountsException;

	/**
	 * Set a general purpose value on the user record.
	 * @param userId
	 * @param key
	 * @param value
	 * @throws DrumlinAccountsException 
	 */
	void setUserValue ( String userId, String key, String value ) throws DrumlinAccountsException;

	/**
	 * Retrieve a general purpose value from the user record.
	 * @param userId
	 * @param key
	 * @return the stored value, or null
	 * @throws DrumlinAccountsException 
	 */
	String getUserValue ( String userId, String key ) throws DrumlinAccountsException;

	/**
	 * Remove a general purpose value from the user record.
	 * @param userId
	 * @param key
	 */
	void removeUservalue ( String userId, String key ) throws DrumlinAccountsException;
}
