/*
 *	Copyright 2006-2013, Rathravane LLC
 *
 *	Licensed under the Apache License, Version 2.0 (the "License");
 *	you may not use this file except in compliance with the License.
 *	You may obtain a copy of the License at
 *	
 *	http://www.apache.org/licenses/LICENSE-2.0
 *	
 *	Unless required by applicable law or agreed to in writing, software
 *	distributed under the License is distributed on an "AS IS" BASIS,
 *	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *	See the License for the specific language governing permissions and
 *	limitations under the License.
 */
package com.rathravane.drumlin.service.framework.routing.playish;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.List;
import java.util.logging.Logger;

import com.rathravane.drumlin.service.framework.context.DrumlinRequestContext;
import com.rathravane.drumlin.service.standards.MimeTypes;
import com.rathravane.till.data.rrStreamTools;
import com.rathravane.till.logging.rrLogSetup;

public class StaticDirHandler implements DrumlinPlayishRouteHandler
{
	public static final String kMaxAge = StaticFileHandler.kMaxAge;

	public StaticDirHandler ( String staticDirInfo )
	{
		// the format of staticDirInfo is "dir;defaultpage"
		final String[] parts = staticDirInfo.split ( ";" );
		if ( parts.length < 1 ) throw new IllegalArgumentException ( "dir[;defaultpage]" );

		if ( parts.length > 1 )
		{
			fDefaultPage = parts[1];
		}
		else
		{
			fDefaultPage = null;
		}
	}

	@Override
	public void handle ( DrumlinRequestContext context, List<String> args )
	{
		final String path = context.request ().getPathInContext ();
		if ( path != null && path.length() > 0 )
		{
			if ( path.contains ( ".." ) )
			{
				log.warning ( "404 [" + path + "] contains parent directory accessor" );
				context.response ().sendError ( 404, path + " was not found on this server." );
			}
			else
			{
				File in = context.getServlet ().findFile ( "." + path );
				if ( ( !in.exists () || !in.isFile () ) && fDefaultPage != null )
				{
					String defIn = in + File.separator + fDefaultPage;
					log.info ( "[" + path + "] does not exist, trying [" + defIn + "]." );
					File defInFile = context.getServlet ().findFile ( defIn );
					if ( defInFile.exists () && defInFile.isFile () )
					{
						log.info ( "[" + defInFile + "] found, using it." );
						in = defInFile;
					}
				}

				log.info ( "Path [" + path + "] ==> [" + in.getAbsolutePath () + "]." );
				final String contentType = mapToContentType ( in.getName () );

				// expiry. currently global.
				final int cacheMaxAge = context.systemSettings ().getInt ( kMaxAge, -1 );
				if ( cacheMaxAge > 0 )
				{
					context.response().writeHeader ( "Cache-Control", "max-age=" + cacheMaxAge, true );
				}
				
				try
				{
					final FileInputStream is = new FileInputStream ( in );
					final OutputStream os = context.response ().getStreamForBinaryResponse ( contentType );
					rrStreamTools.copyStream ( is, os );
				}
				catch ( FileNotFoundException e )
				{
					log.warning ( "404 [" + path + "]==>[" + path + "] (" + in.getAbsolutePath () + ")" );
					context.response ().sendError ( 404, path + " was not found on this server." );
				}
				catch ( IOException e )
				{
					log.warning ( "500 [" + in.getAbsolutePath () + "]: " + e.getMessage () );
					context.response ().sendError ( 500, e.getMessage () );
				}
			}
		}
		else
		{
			throw new IllegalArgumentException ( "no path provided" );
		}
	}

	private final String fDefaultPage;

	private static final Logger log = rrLogSetup.getLog ( StaticDirHandler.class );

	static final HashMap<String,String> sfContentTypes = new HashMap<String,String> ();
	static
	{
		sfContentTypes.put ( "css", MimeTypes.kCss );

		sfContentTypes.put ( "jpg", MimeTypes.kImageJpg );
		sfContentTypes.put ( "gif", MimeTypes.kImageGif );
		sfContentTypes.put ( "png", MimeTypes.kImagePng );
		sfContentTypes.put ( "ico", MimeTypes.kImageIco );

		sfContentTypes.put ( "htm", MimeTypes.kHtml );
		sfContentTypes.put ( "html", MimeTypes.kHtml );

		sfContentTypes.put ( "js", MimeTypes.kAppJavascript );

		sfContentTypes.put ( "eot", MimeTypes.kFontEot );
		sfContentTypes.put ( "woff", MimeTypes.kFontWoff );
		sfContentTypes.put ( "otf", MimeTypes.kFontOtf );
		sfContentTypes.put ( "ttf", MimeTypes.kFontTtf );
		sfContentTypes.put ( "svg", MimeTypes.kSvg );
	}

	public static String mapToContentType ( String name )
	{
		final int dot = name.lastIndexOf ( "." );
		if ( dot != -1 )
		{
			name = name.substring ( dot + 1 );
		}
		String result = sfContentTypes.get ( name );
		if ( result == null )
		{
			log.warning ( "Unknown content type [" + name + "]. Sending text/plain. (See " + StaticDirHandler.class.getSimpleName () + "::sfContentTypes)" );
			result = "text/plain";
		}
		return result;
	}

	@Override
	public boolean actionMatches(String fullPath)
	{
		return false;
	}
}

