/*
 *	Copyright 2006-2013, Rathravane LLC
 *
 *	Licensed under the Apache License, Version 2.0 (the "License");
 *	you may not use this file except in compliance with the License.
 *	You may obtain a copy of the License at
 *	
 *	http://www.apache.org/licenses/LICENSE-2.0
 *	
 *	Unless required by applicable law or agreed to in writing, software
 *	distributed under the License is distributed on an "AS IS" BASIS,
 *	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *	See the License for the specific language governing permissions and
 *	limitations under the License.
 */
package com.rathravane.drumlin.service.standalone;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;

import winstone.Launcher;

import com.rathravane.till.console.cmdLinePrefs;
import com.rathravane.till.console.rrCmdLineParser;
import com.rathravane.till.console.rrConsole;
import com.rathravane.till.nv.rrNvReadable;
import com.rathravane.till.nv.rrNvWriteable;

/**
 * A standalone Drumlin server built on Winstone. To use this class, derive a class,
 * instantiate it, and call "runFromMain".
 *  
 * @author peter@rathravane.com
 *
 */
public class DrumlinStandaloneServer extends rrConsole
{
	public static final String kSetting_WebRoot = "webroot";
	public static final String kSetting_PortNumber = "port";

	public static final String kSetting_HttpsPort = "httpsPort";
	public static final String kSetting_HttpsKeystore = "httpsKeyStore";
	public static final String kSetting_HttpsKeystorePassword = "httpsKeyStorePassword";

	private static final int kSleepDuringLoopMs = 250;

	/**
	 * Set the defaults for the server.
	 */
	@Override
	protected void setupDefaults ( rrNvWriteable pt )
	{
		pt.set ( kSetting_WebRoot, "." );
		pt.set ( kSetting_PortNumber, "8080" );

		pt.set ( kSetting_HttpsPort, -1 );
	}

	/**
	 * Setup the options for the base class console system.
	 */
	@Override
	protected void setupOptions ( rrCmdLineParser p )
	{
		p.registerOptionWithValue ( kSetting_WebRoot );
		p.registerOptionWithValue ( kSetting_PortNumber, "p", null, null );

		p.registerOptionWithValue ( kSetting_HttpsPort );
		p.registerOptionWithValue ( kSetting_HttpsKeystore );
		p.registerOptionWithValue ( kSetting_HttpsKeystorePassword );
	}

	/**
	 * Setup and run the Winstone server, then wait in a loop until the
	 * server instance reports that it's stopped.
	 */
	@Override
	protected looper init ( rrNvReadable p, cmdLinePrefs clp ) throws rrNvReadable.missingReqdSetting, rrNvReadable.invalidSettingValue, startupFailureException
	{
		fCommandLineSettings = p;

		return new looper ()
		{
			@Override
			public boolean setup ( rrNvReadable p, cmdLinePrefs clp )
			{
				boolean result = false;
				try
				{
					fWebRoot = p.getString ( kSetting_WebRoot );
					fPort = p.getInt ( kSetting_PortNumber );

					fHttpsPort = p.getInt ( kSetting_HttpsPort, -1 );
					fHttpsKeyStore = p.getString ( kSetting_HttpsKeystore, null );
					fHttpsKeyStorePwd = p.getString ( kSetting_HttpsKeystorePassword, null );

					result = serverStart ();
				}
				catch ( rrNvReadable.missingReqdSetting e )
				{
					log.severe ( e.getMessage () );
				}
				return result;
			}

			@Override
			public boolean loop ( rrNvReadable p )
			{
				try
				{
					Thread.sleep ( kSleepDuringLoopMs );
				}
				catch ( InterruptedException e )
				{
					// ignore
				}
				return serverIsRunning ();
			}

			@Override
			public void teardown ( rrNvReadable p )
			{
				serverStop ();
			}
		};
	}

	/**
	 * Start the web server
	 * @return true if the server starts.
	 */
	public boolean serverStart ()
	{
		log.fine ( "Web service starting on port " + fPort + ", from configuration [" + fWebRoot + "/web.xml]." );

		final Map<String,String> winstoneArgs = new HashMap<String,String> ();
		winstoneArgs.put ( "webroot", fWebRoot );
		winstoneArgs.put ( "httpPort", "" + fPort );
		winstoneArgs.put ( "ajp13Port", "-1" );

		if ( fHttpsPort != -1 )
		{
			winstoneArgs.put ( "httpsPort", "" + fHttpsPort );
		}
		if ( fHttpsKeyStore != null )
		{
			winstoneArgs.put ( "httpsKeyStore", fHttpsKeyStore );
		}
		if ( fHttpsKeyStorePwd != null )
		{
			winstoneArgs.put ( "httpsKeyStorePassword", fHttpsKeyStorePwd );
		}

		try
		{
			Launcher.initLogger(winstoneArgs);
			fWinstone = new Launcher(winstoneArgs);
			return true;
		}
		catch ( IOException e )
		{
			log.severe ( "Couldn't start service. " + e.getMessage () );
			return false;
		}
	}

	/**
	 * Stop the server.
	 */
	public void serverStop ()
	{
		if ( fWinstone != null )
		{
			fWinstone.shutdown ();
			fWinstone = null;
		}
	}

	/**
	 * Return true if the server is running.
	 * @return true if the server is running
	 */
	public boolean serverIsRunning ()
	{
		return fWinstone != null && fWinstone.isRunning ();
	}

	protected DrumlinStandaloneServer ()
	{
		fWebRoot = null;
		fPort = -1;
		fHttpsPort = -1;
		fWinstone = null;
		fCommandLineSettings = null;
		fHttpsKeyStore = null;
		fHttpsKeyStorePwd = null;
	}

	protected rrNvReadable getCommandLineSettings ()
	{
		return fCommandLineSettings;
	}

	private int fPort;
	private String fWebRoot;
	private Launcher fWinstone;
	private rrNvReadable fCommandLineSettings;

	private int fHttpsPort;
	private String fHttpsKeyStore;
	private String fHttpsKeyStorePwd;

	private static final Logger log = Logger.getLogger ( DrumlinStandaloneServer.class.getName () );
}
