package com.rathravane.drumlin.app.accounts;

import java.util.Collection;
import java.util.HashMap;
import java.util.concurrent.TimeUnit;

import junit.framework.TestCase;

import org.junit.Test;

import com.rathravane.till.time.clock;

public class drumlinAccountMgrTest extends TestCase
{
	@Test
	public void testSignonWithPasswordOk () throws DrumlinAccountsException
	{
		final DrumlinAccountPersistence<user,account> dap = getModel ();
		final DrumlinAccountMgr<user,account> am = new DrumlinAccountMgr<user,account> ( dap );

		am.createAccount ( "a1" );
		am.enableAccount ( "a1", true );

		am.createUser ( "a1", "u1", "u1@a1.example.com" );
		am.enableUser ( "u1", true );
		am.setPassword ( "u1", "pwd" );

		final user u = am.authenticateUserWithPassword ( "u1", "pwd" );
		assertNotNull ( u );
		assertEquals ( "a1", u.fAcctId );
	}

	@Test
	public void testSignonWithBadPassword () throws DrumlinAccountsException
	{
		final DrumlinAccountPersistence<user,account> dap = getModel ();
		final DrumlinAccountMgr<user,account> am = new DrumlinAccountMgr<user,account> ( dap );

		am.createAccount ( "a1" );
		am.enableAccount ( "a1", true );

		am.createUser ( "a1", "u1", "u1@a1.example.com" );
		am.enableUser ( "u1", true );
		am.setPassword ( "u1", "pwd" );

		final user u = am.authenticateUserWithPassword ( "u1", "bad" );
		assertNull ( u );
	}

	@Test
	public void testSignonWithNoPasswordSet () throws DrumlinAccountsException
	{
		final DrumlinAccountPersistence<user,account> dap = getModel ();
		final DrumlinAccountMgr<user,account> am = new DrumlinAccountMgr<user,account> ( dap );

		am.createAccount ( "a1" );
		am.enableAccount ( "a1", true );

		am.createUser ( "a1", "u1", "u1@a1.example.com" );
		am.enableUser ( "u1", true );

		final user u = am.authenticateUserWithPassword ( "u1", "bad" );
		assertNull ( u );
	}

	@Test
	public void testSignonWithUserDisabled () throws DrumlinAccountsException
	{
		final DrumlinAccountPersistence<user,account> dap = getModel ();
		final DrumlinAccountMgr<user,account> am = new DrumlinAccountMgr<user,account> ( dap );

		am.createAccount ( "a1" );
		am.enableAccount ( "a1", true );

		am.createUser ( "a1", "u1", "u1@a1.example.com" );
		am.enableUser ( "u1", false );
		am.setPassword ( "u1", "pwd" );

		final user u = am.authenticateUserWithPassword ( "u1", "pwd" );
		assertNull ( u );
	}

	@Test
	public void testSignonWithAccountDisabled () throws DrumlinAccountsException
	{
		final DrumlinAccountPersistence<user,account> dap = getModel ();
		final DrumlinAccountMgr<user,account> am = new DrumlinAccountMgr<user,account> ( dap );

		am.createAccount ( "a1" );
		am.enableAccount ( "a1", false );
		assertTrue ( am.accountExists ( "a1" ) );

		am.createUser ( "a1", "u1", "u1@a1.example.com" );
		am.enableUser ( "u1", true );
		am.setPassword ( "u1", "pwd" );
		assertTrue ( am.userExists ( "u1" ) );

		final user u = am.authenticateUserWithPassword ( "u1", "pwd" );
		assertNull ( u );
	}

	@Test
	public void testPasswordReset () throws DrumlinAccountsException
	{
		final DrumlinAccountPersistence<user,account> dap = getModel ();
		final DrumlinAccountMgr<user,account> am = new DrumlinAccountMgr<user,account> ( dap );

		am.createAccount ( "a1" );
		am.enableAccount ( "a1", true );
		assertTrue ( am.accountExists ( "a1" ) );

		am.createUser ( "a1", "u1", "u1@a1.example.com" );
		am.enableUser ( "u1", true );
		am.setPassword ( "u1", "pwd" );
		assertTrue ( am.userExists ( "u1" ) );

		user u = am.authenticateUserWithPassword ( "u1", "pwd" );
		assertNotNull ( u );

		final String token = am.requestPasswordReset ( "u1", 60 * 60 * 24, "" + clock.now () );

		// password should not change yet
		u = am.authenticateUserWithPassword ( "u1", "pwd" );
		assertNotNull ( u );

		am.completePasswordReset ( token, "new" );

		// old password fails
		u = am.authenticateUserWithPassword ( "u1", "pwd" );
		assertNull ( u );

		// new password works
		u = am.authenticateUserWithPassword ( "u1", "new" );
		assertNotNull ( u );
	}

	@Test
	public void testPasswordResetFabricatedToken () throws DrumlinAccountsException
	{
		final DrumlinAccountPersistence<user,account> dap = getModel ();
		final DrumlinAccountMgr<user,account> am = new DrumlinAccountMgr<user,account> ( dap );

		am.createAccount ( "a1" );
		am.enableAccount ( "a1", true );
		assertTrue ( am.accountExists ( "a1" ) );

		am.createUser ( "a1", "u1", "u1@a1.example.com" );
		am.enableUser ( "u1", true );
		am.setPassword ( "u1", "pwd" );
		assertTrue ( am.userExists ( "u1" ) );

		user u = am.authenticateUserWithPassword ( "u1", "pwd" );
		assertNotNull ( u );

		am.completePasswordReset ( "bogus token", "new" );

		// old password still works
		u = am.authenticateUserWithPassword ( "u1", "pwd" );
		assertNotNull ( u );

		// new password fails
		u = am.authenticateUserWithPassword ( "u1", "new" );
		assertNull ( u );
	}

	@Test
	public void testPasswordResetExpiredToken () throws DrumlinAccountsException
	{
		// setup a test clock
		final clock.testClock tc = clock.useNewTestClock ();
		
		final DrumlinAccountPersistence<user,account> dap = getModel ();
		final DrumlinAccountMgr<user,account> am = new DrumlinAccountMgr<user,account> ( dap );

		am.createAccount ( "a1" );
		am.enableAccount ( "a1", true );
		assertTrue ( am.accountExists ( "a1" ) );

		am.createUser ( "a1", "u1", "u1@a1.example.com" );
		am.setPassword ( "u1", "pwd" );
		am.enableUser ( "u1", true );
		assertTrue ( am.userExists ( "u1" ) );

		user u = am.authenticateUserWithPassword ( "u1", "pwd" );
		assertNotNull ( u );

		// token that expires in 8 hrs
		final String token = am.requestPasswordReset ( "u1", 8 * 60 * 60, "" + clock.now () );

		// advance the clock one day
		tc.add ( 24, TimeUnit.HOURS );

		// call the password reset
		final boolean reset = ( null != am.completePasswordReset ( token, "new" ) );
		assertFalse ( reset );

		// old password still works
		u = am.authenticateUserWithPassword ( "u1", "pwd" );
		assertNotNull ( u );

		// new password fails
		u = am.authenticateUserWithPassword ( "u1", "new" );
		assertNull ( u );
	}

	private static class userTagInfo
	{
		public userTagInfo ( String userId, String tag, String tagType, long expiresMs ) { fUserId=userId; fTag=tag; fTagType=tagType; fExpires=expiresMs; };
		public final String fUserId;
		public final String fTag;
		public final String fTagType;
		public final long fExpires;
	};

	private static class user
	{
		public boolean fEnabled = false;
		public String fAcctId = null;
		public String fSalt = null;
		public String fHashedPwd = null;
		public HashMap<String,userTagInfo> fTags = new HashMap<String,userTagInfo> ();
	}

	private static class account
	{
		public boolean fEnabled = false;
	}

	private DrumlinAccountPersistence<user,account> getModel ()
	{
		return new DrumlinAccountPersistence<user,account>()
		{
			private final HashMap<String,user> fUserMap = new HashMap<String,user> ();
			private final HashMap<String,account> fAcctMap = new HashMap<String,account> ();
			private final HashMap<String,userTagInfo> fTags = new HashMap<String,userTagInfo> ();

			@Override
			public String getAppSignature ()
			{
				return "drumlinUnitTesting";
			}

			@Override
			public void close ()
			{
				fUserMap.clear ();
			}

			@Override
			public user createUser ( String acctId, String uid, String userEmail ) throws DrumlinAccountsException
			{
				if ( userExists ( uid ) ) throw new DrumlinAccountsException ( "user exists" );
				final user u = new user ();
				u.fAcctId = acctId;
				fUserMap.put ( uid, u );
				return u;
			}

			@Override
			public void deleteUser ( String uid ) throws DrumlinAccountsException
			{
				if ( userExists ( uid ) )
				{
					fUserMap.remove ( uid );
				}
			}

			@Override
			public user loadUser ( String uid ) throws DrumlinAccountsException
			{
				if ( !userExists ( uid ) ) throw new DrumlinAccountsException ( "no such user" );
				return fUserMap.get ( uid );
			}

			@Override
			public boolean userExists ( String uid ) throws DrumlinAccountsException
			{
				return fUserMap.containsKey ( uid );
			}

			@Override
			public void enableUser ( String id, boolean enabled ) throws DrumlinAccountsException
			{
				loadUser ( id ).fEnabled = enabled;
			}

			@Override
			public boolean isUserEnabled ( String userId ) throws DrumlinAccountsException
			{
				return loadUser ( userId ).fEnabled;
			}

			@Override
			public String getAccountIdForUser ( String userId ) throws DrumlinAccountsException
			{
				return loadUser ( userId ).fAcctId;
			}

			@Override
			public void storePasswordInfo ( String userId, String salt, String hashedPassword ) throws DrumlinAccountsException
			{
				final user u = loadUser ( userId );
				u.fSalt = salt;
				u.fHashedPwd = hashedPassword;
			}

			@Override
			public String[] getPasswordInfo ( String userId ) throws DrumlinAccountsException
			{
				final user u = loadUser ( userId );
				return new String[] { u.fSalt, u.fHashedPwd };
			}

			@Override
			public void storeRequestTag ( String userId, String tagType, String tag, long expiresAtEpochTime ) throws DrumlinAccountsException
			{
				final user u = loadUser ( userId );
				final userTagInfo uti = new userTagInfo ( userId, tag, tagType, expiresAtEpochTime );
				u.fTags.put ( tagType, uti );
				fTags.put ( tag, uti );
			}

			@Override
			public tagInfo getRequestTagByTag ( String tag, String tagType ) throws DrumlinAccountsException
			{
				final userTagInfo uti = fTags.get ( tag );
				if ( uti == null ) return null;
				return new tagInfo ( tag, uti.fUserId, uti.fExpires );
			}

			@Override
			public tagInfo getRequestTagByUser ( String userId, String tagType ) throws DrumlinAccountsException
			{
				final user u = loadUser ( userId );
				for ( userTagInfo uti : u.fTags.values () )
				{
					if ( uti.fTagType.equals ( tagType ) )
					{
						return new tagInfo ( uti.fTag, uti.fUserId, uti.fExpires );
					}
				}
				return null;
			}

			@Override
			public void deleteRequestByTag ( String tag, String tagType ) throws DrumlinAccountsException
			{
				final userTagInfo uti = fTags.get ( tag );
				if ( uti != null && uti.fTagType.equals ( tagType ) )
				{
					fTags.remove ( tag );
				}
			}

			@Override
			public void deleteRequestTagOnUser ( String userId, String tag, String tagType ) throws DrumlinAccountsException
			{
				final user u = loadUser ( userId );
				final userTagInfo uti = u.fTags.get ( tagType );
				if ( uti != null && uti.fTag.equals ( tag ) )
				{
					u.fTags.remove ( tag );
				}
			}

			@Override
			public account createAccount ( String id ) throws DrumlinAccountsException
			{
				if ( accountExists ( id ) ) throw new DrumlinAccountsException ( "account exists" );

				final account a = new account ();
				fAcctMap.put ( id, a );
				return a;
			}

			@Override
			public account loadAccount ( String id ) throws DrumlinAccountsException
			{
				if ( !accountExists ( id ) ) throw new DrumlinAccountsException ( "account does not exist" );
				return fAcctMap.get ( id );
			}

			@Override
			public boolean accountExists ( String id ) throws DrumlinAccountsException
			{
				return fAcctMap.containsKey ( id );
			}

			@Override
			public void enableAccount ( String id, boolean enabled ) throws DrumlinAccountsException
			{
				loadAccount ( id ).fEnabled = enabled;
			}

			@Override
			public boolean isAccountEnabled ( String id ) throws DrumlinAccountsException
			{
				return loadAccount ( id ).fEnabled;
			}

			@Override
			public DrumlinApiKey loadApiKey ( String apiKey ) throws DrumlinAccountsException
			{
				throw new IllegalStateException ( "not implemented" );
			}

			@Override
			public void storeApiKey ( DrumlinApiKey apiKey ) throws DrumlinAccountsException
			{
				throw new IllegalStateException ( "not implemented" );
			}

			@Override
			public void setApiSecret ( String apiKey, String apiSecret ) throws DrumlinAccountsException
			{
				throw new IllegalStateException ( "not implemented" );
			}

			@Override
			public void deleteApiKey ( DrumlinApiKey apiKey ) throws DrumlinAccountsException
			{
				throw new IllegalStateException ( "not implemented" );
			}

			@Override
			public Collection<DrumlinApiKey> loadApiKeysForUser ( String userId ) throws DrumlinAccountsException
			{
				throw new IllegalStateException ( "not implemented" );
			}

			@Override
			public void setUserValue ( String userId, String key, String value ) throws DrumlinAccountsException
			{
				throw new IllegalStateException ( "not implemented" );
			}

			@Override
			public String getUserValue ( String userId, String key ) throws DrumlinAccountsException
			{
				throw new IllegalStateException ( "not implemented" );
			}

			@Override
			public void removeUservalue ( String userId, String key ) throws DrumlinAccountsException
			{
				throw new IllegalStateException ( "not implemented" );
			}
		};
	}
}
